import asyncio
import sys
import types
from pathlib import Path

from panther import status
from panther._load_configs import *
from panther._utils import clean_traceback_message, http_response, read_body
from panther.configs import config
from panther.exceptions import APIException
from panther.middlewares.monitoring import Middleware as MonitoringMiddleware
from panther.request import Request
from panther.response import Response
from panther.routings import collect_path_variables, find_endpoint

""" We can't import logger on the top cause it needs config['base_dir'] ans its fill in __init__ """


class Panther:

    def __init__(self, name, configs=None):
        from panther.logger import logger

        self._configs = configs
        config['base_dir'] = Path(name).resolve().parent
        if sys.version_info.minor < 11:
            logger.warning('Use Python Version 3.11+ For Better Performance.')

        try:
            self.load_configs()
        except TypeError:
            exit()

    def load_configs(self) -> None:
        from panther.logger import logger
        logger.debug(f'Base directory: {config["base_dir"]}')

        # Check & Read The Configs File
        self.configs = load_configs_file(self._configs)

        # Put Variables In "config" (Careful about the ordering)
        config['secret_key'] = load_secret_key(self.configs)
        config['monitoring'] = load_monitoring(self.configs)
        config['log_queries'] = load_log_queries(self.configs)
        config['throttling'] = load_throttling(self.configs)
        config['default_cache_exp'] = load_default_cache_exp(self.configs)
        config['middlewares'] = load_middlewares(self.configs)
        config['reversed_middlewares'] = config['middlewares'][::-1]
        config['user_model'] = load_user_model(self.configs)
        config['authentication'] = load_authentication_class(self.configs)
        config['jwt_config'] = load_jwt_config(self.configs)
        config['models'] = collect_all_models()

        # Load URLs should be the last call in load_configs,
        #   because it will read all files and load them.
        config['urls'] = load_urls(self.configs)
        config['urls']['_panel'] = load_panel_urls()

        if config['monitoring']:
            logger.info('Run "panther monitor" in another session for Monitoring.')

    async def __call__(self, scope, receive, send) -> None:
        """
        We Used Python3.11+ For asyncio.TaskGroup()
        1.
            async with asyncio.TaskGroup() as tg:
                tg.create_task(self.run(scope, receive, send))
        2.
            await self.run(scope, receive, send)
        3.
            async with anyio.create_task_group() as task_group:
                task_group.start_soon(self.run, scope, receive, send)
                await anyio.to_thread.run_sync(self.run, scope, receive, send)
        4.
            with ProcessPoolExecutor() as e:
                e.submit(self.run, scope, receive, send)
        """
        if sys.version_info.minor >= 11:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(self.run(scope, receive, send))
        else:
            await self.run(scope, receive, send)

    async def run(self, scope, receive, send):
        from panther.logger import logger

        request = Request(scope=scope)

        # Monitoring Middleware
        monitoring_middleware = None
        if config['monitoring']:
            monitoring_middleware = MonitoringMiddleware()
            await monitoring_middleware.before(request=request)

        # Read Body & Create Request
        body = await read_body(receive)
        request.set_body(body)

        # Find Endpoint
        endpoint, found_path = find_endpoint(path=request.path)
        path_variables: dict = collect_path_variables(request_path=request.path, found_path=found_path)

        if endpoint is None:
            return await http_response(
                send, status_code=status.HTTP_404_NOT_FOUND, monitoring=monitoring_middleware, exception=True,
            )

        try:  # They Both(middleware.before() & _endpoint()) Have The Same Exception (APIException)
            # Call 'Before' Middlewares
            for middleware in config['middlewares']:
                request = await middleware.before(request=request)

            # Function
            if isinstance(endpoint, types.FunctionType):
                # Function Doesn't Have @API Decorator
                if not hasattr(endpoint, '__wrapped__'):
                    logger.critical(f'You may have forgotten to use @API on the {endpoint.__name__}()')
                    return await http_response(
                        send,
                        status_code=status.HTTP_501_NOT_IMPLEMENTED,
                        monitoring=monitoring_middleware,
                        exception=True,
                    )

                # Declare Endpoint
                _endpoint = endpoint

            # Class
            else:
                from panther.app import GenericAPI

                if not issubclass(endpoint, GenericAPI):
                    logger.critical(f'You may have forgotten to inherit from GenericAPI on the {endpoint.__name__}()')
                    return await http_response(
                        send,
                        status_code=status.HTTP_501_NOT_IMPLEMENTED,
                        monitoring=monitoring_middleware,
                        exception=True,
                    )
                # Declare Endpoint
                _endpoint = endpoint.call_method

            # Call Endpoint
            response = await _endpoint(request=request, **path_variables)

        except APIException as e:
            response = self.handle_exceptions(e)
        except Exception as e:  # noqa: BLE001
            # Every unhandled exception in Panther or code will catch here
            exception = clean_traceback_message(exception=e)
            logger.critical(exception)
            return await http_response(
                send,
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                monitoring=monitoring_middleware,
                exception=True,
            )

        # Call 'After' Middleware
        for middleware in config['reversed_middlewares']:
            try:
                response = await middleware.after(response=response)
            except APIException as e:  # noqa: PERF203
                response = self.handle_exceptions(e)

        await http_response(
            send,
            status_code=response.status_code,
            monitoring=monitoring_middleware,
            headers=response.headers,
            body=response.body,
        )

    @classmethod
    def handle_exceptions(cls, e, /) -> Response:
        return Response(
            data=e.detail if isinstance(e.detail, dict) else {'detail': e.detail},
            status_code=e.status_code,
        )



import importlib
import re
from traceback import TracebackException

import orjson as json

from panther import status
from panther.file_handler import File
from panther.logger import logger


async def read_body(receive) -> bytes:
    """Read and return the entire body from an incoming ASGI message."""
    body = b''
    more_body = True
    while more_body:
        message = await receive()
        # {'type': 'lifespan.startup'}
        body += message.get('body', b'')
        more_body = message.get('more_body', False)
    return body


async def _http_response_start(send, /, headers: dict, status_code: int):
    bytes_headers = [[k.encode(), v.encode()] for k, v in (headers or {}).items()]
    await send({
        'type': 'http.response.start',
        'status': status_code,
        'headers': bytes_headers,
    })


async def _http_response_body(send, /, body: any = None):
    if body is None:
        await send({'type': 'http.response.body'})
    else:
        await send({'type': 'http.response.body', 'body': body})


async def http_response(
        send,
        /,
        *,
        status_code: int,
        monitoring=None,  # type: MonitoringMiddleware | None
        headers: dict = None,
        body: bytes = None,
        exception: bool = False,
):
    if exception:
        body = json.dumps({'detail': status.status_text[status_code]})
    elif status_code == status.HTTP_204_NO_CONTENT or body == b'null':
        body = None

    if monitoring is not None:
        await monitoring.after(status_code=status_code)

    # TODO: Should we send 'access-control-allow-origin' in any case
    await _http_response_start(send, headers=headers, status_code=status_code)
    await _http_response_body(send, body=body)


def import_class(dotted_path: str, /):
    """
    Example:
    -------
        Input: panther.db.models.User
        Output: User (The Class)
    """
    path, name = dotted_path.rsplit('.', 1)
    module = importlib.import_module(path)
    return getattr(module, name)


def read_multipart_form_data(boundary: str, body: bytes) -> dict:
    """
    ----------------------------449529189836774544725855
    \r\nContent-Disposition: form-data; name="name"\r\n\r\nali\r\n
    ----------------------------449529189836774544725855
    \r\nContent-Disposition: form-data; name="image"; filename="ali.txt"\r\nContent-Type: text/plain\r\n\r\nHello\n\r\n
    ----------------------------449529189836774544725855
    \r\nContent-Disposition: form-data; name="age"\r\n\r\n12\r\n
    ----------------------------449529189836774544725855
    --\r\n
    """

    boundary = b'--' + boundary.encode()
    new_line = b'\r\n' if body[-2:] == b'\r\n' else b'\n'

    field_pattern = rb'(Content-Disposition: form-data; name=")(.*)("' + 2 * new_line + b')(.*)'
    file_pattern = rb'(Content-Disposition: form-data; name=")(.*)("; filename=")(.*)("' + new_line + b'Content-Type: )(.*)'

    data = dict()
    for row in body.split(boundary):

        row = row.removeprefix(new_line).removesuffix(new_line)
        if row == b'' or row == b'--':
            continue

        if match := re.match(pattern=field_pattern, string=row):
            _, field_name, _, value = match.groups()
            data[field_name.decode('utf-8')] = value.decode('utf-8')

        else:
            file_meta_data, value = row.split(2 * new_line, 1)
            if match := re.match(pattern=file_pattern, string=file_meta_data):
                _, field_name, _, file_name, _, content_type = match.groups()
                file = File(
                    file_name=file_name.decode('utf-8'),
                    content_type=content_type.decode('utf-8'),
                    file=value,
                )
                data[field_name.decode('utf-8')] = file
            else:
                logger.error('Unrecognized Pattern')

    return data


def is_function_async(func) -> bool:
    """
        sync result is 0 --> False
        async result is 128 --> True
    """
    return bool(func.__code__.co_flags & (1 << 7))


def clean_traceback_message(exception) -> str:
    """
    We are ignoring packages traceback message
    """
    tb = TracebackException(type(exception), exception, exception.__traceback__)
    stack = tb.stack.copy()
    for t in stack:
        if t.filename.find('site-packages') != -1:
            tb.stack.remove(t)
    return f'{exception}\n' + ''.join(tb.format(chain=False))



import functools
from datetime import datetime, timedelta

from orjson.orjson import JSONDecodeError
from pydantic import ValidationError

from panther import status
from panther._utils import is_function_async
from panther.caching import cache_key, get_cached_response_data, set_cache_response
from panther.configs import config
from panther.exceptions import (
    APIException,
    AuthorizationException,
    InvalidPathVariableException,
    JsonDecodeException,
    MethodNotAllowed,
    ThrottlingException,
)
from panther.logger import logger
from panther.request import Request
from panther.response import IterableDataTypes, Response
from panther.throttling import Throttling, throttling_storage
from panther.utils import round_datetime

__all__ = ('API', 'GenericAPI')


class API:
    def __init__(
            self,
            *,
            input_model=None,
            output_model=None,
            auth: bool = False,
            permissions: list | None = None,
            throttling: Throttling = None,
            cache: bool = False,
            cache_exp_time: timedelta | int | None = None,
    ):
        self.input_model = input_model
        self.output_model = output_model
        self.auth = auth
        self.permissions = permissions or []
        self.throttling = throttling
        self.cache = cache
        self.cache_exp_time = cache_exp_time
        self.request: Request | None = None

    def __call__(self, func):
        @functools.wraps(func)
        async def wrapper(request: Request, **path_variables):
            self.request: Request = request  # noqa: Non-self attribute could not be type hinted

            # 1. Authentication
            self.handle_authentications()

            # 2. Throttling
            self.handle_throttling()

            # 3. Permissions
            self.handle_permissions()

            # 4. Validate Input
            if self.request.method in ['POST', 'PUT', 'PATCH']:
                self.set_validated_input()

            # 5. Validate Path Variables
            self.validate_path_variables(func, path_variables)

            # 6. Get Cached Response
            if self.cache and self.request.method == 'GET':
                if cached := get_cached_response_data(request=self.request):
                    return Response(data=cached.data, status_code=cached.status_code)

            # 7. Put Request In kwargs (If User Wants It)
            kwargs = path_variables
            if req_arg := [k for k, v in func.__annotations__.items() if v == Request]:
                kwargs[req_arg[0]] = self.request

            # 8. Call Endpoint
            if is_function_async(func):
                response = await func(**kwargs)
            else:
                response = func(**kwargs)

            # 9. Clean Output
            if not isinstance(response, Response):
                response = Response(data=response)
            data = self.serialize_response_data(data=response._data)  # noqa: SLF001
            response.set_data(data)

            # 10. Set New Response To Cache
            if self.cache and self.request.method == 'GET':
                set_cache_response(request=self.request, response=response, cache_exp_time=self.cache_exp_time)

            # 11. Warning CacheExpTime
            if self.cache_exp_time and self.cache is False:
                logger.warning('"cache_exp_time" won\'t work while "cache" is False')

            return response

        return wrapper

    def handle_authentications(self) -> None:
        auth_class = config['authentication']
        if self.auth:
            if not auth_class:
                raise TypeError('"AUTHENTICATION" has not been set in core/configs')
            user = auth_class.authentication(self.request)
            self.request.set_user(user=user)

    def handle_throttling(self) -> None:
        if throttling := self.throttling or config['throttling']:
            key = cache_key(self.request)
            time = round_datetime(datetime.now(), throttling.duration)
            throttling_key = f'{time}-{key}'
            if throttling_storage[throttling_key] > throttling.rate:
                raise ThrottlingException

            throttling_storage[throttling_key] += 1

    def handle_permissions(self) -> None:
        for perm in self.permissions:
            if type(perm.authorization).__name__ != 'method':
                logger.error(f'{perm.__name__}.authorization should be "classmethod"')
                continue
            if perm.authorization(request=self.request) is False:
                raise AuthorizationException

    def set_validated_input(self):
        if self.input_model:
            validated_data = self.validate_input(model=self.input_model, request=self.request)
            self.request.set_validated_data(validated_data)

    @classmethod
    def validate_input(cls, model, request: Request):
        try:
            return model(**request.pure_data)
        except ValidationError as validation_error:
            error = {'.'.join(loc for loc in e['loc']): e['msg'] for e in validation_error.errors()}
            raise APIException(status_code=status.HTTP_400_BAD_REQUEST, detail=error)
        except JSONDecodeError:
            raise JsonDecodeException

    def serialize_response_data(self, data):
        """
        We serializer the response here instead of response.py file
        because we don't have access to the "output_model" there.
        """

        # None or Unchanged
        if data is None or self.output_model is None:
            return data

        return self.serialize_with_output_model(data)

    def serialize_with_output_model(self, data: any):
        # Dict
        if isinstance(data, dict):
            return self.output_model(**data).model_dump()

        # Iterable
        if isinstance(data, IterableDataTypes):
            return [self.serialize_with_output_model(d) for d in data]

        # Str | Bool
        raise TypeError('Type of Response data is not match with output_model. '
                        '\n*hint: You may want to pass None to output_model')

    @staticmethod
    def validate_path_variables(func, request_path_variables: dict):
        for name, value in request_path_variables.items():
            for variable_name, variable_type in func.__annotations__.items():
                if name == variable_name:
                    # Check the type and convert the value
                    if variable_type is bool:
                        request_path_variables[name] = value.lower() not in ['false', '0']

                    elif variable_type is int:
                        try:
                            request_path_variables[name] = int(value)
                        except ValueError:
                            raise InvalidPathVariableException(value=value, variable_type=variable_type)
                    break


class GenericAPI:
    input_model = None
    output_model = None
    auth: bool = False
    permissions: list | None = None
    throttling: Throttling = None
    cache: bool = False
    cache_exp_time: timedelta | int | None = None

    async def get(self, *args, **kwargs):
        raise MethodNotAllowed

    async def post(self, *args, **kwargs):
        raise MethodNotAllowed

    async def put(self, *args, **kwargs):
        raise MethodNotAllowed

    async def patch(self, *args, **kwargs):
        raise MethodNotAllowed

    async def delete(self, *args, **kwargs):
        raise MethodNotAllowed

    @classmethod
    async def call_method(cls, *args, **kwargs):
        match kwargs['request'].method:
            case 'GET':
                func = cls().get
            case 'POST':
                func = cls().post
            case 'PUT':
                func = cls().put
            case 'PATCH':
                func = cls().patch
            case 'DELETE':
                func = cls().delete

        return await API(
            input_model=cls.input_model,
            output_model=cls.output_model,
            auth=cls.auth,
            permissions=cls.permissions,
            throttling=cls.throttling,
            cache=cls.cache,
            cache_exp_time=cls.cache_exp_time,
        )(func)(*args, **kwargs)



from abc import abstractmethod
from datetime import datetime

from jose import JWTError, jwt

from panther.configs import config
from panther.db.models import BaseUser
from panther.exceptions import AuthenticationException
from panther.logger import logger
from panther.request import Request


class BaseAuthentication:
    @classmethod
    @abstractmethod
    def authentication(cls, request: Request):
        """
        Return User Instance
        """
        raise cls.exception(f'{cls.__name__}.authentication() is not implemented.')

    @staticmethod
    def exception(message: str, /):
        logger.error(f'Authentication Error: "{message}"')
        return AuthenticationException


class JWTAuthentication(BaseAuthentication):
    model = BaseUser
    keyword = 'Bearer'
    algorithm = 'HS256'
    HTTP_HEADER_ENCODING = 'iso-8859-1'  # Header encoding (see RFC5987)

    @classmethod
    def get_authorization_header(cls, request: Request):
        auth = request.headers.authorization

        if auth is None:
            raise cls.exception('Authorization is required')

        if isinstance(auth, str):
            auth = auth.encode(JWTAuthentication.HTTP_HEADER_ENCODING)
        return auth

    @classmethod
    def authentication(cls, request: Request):
        auth = cls.get_authorization_header(request).split()
        if not auth or auth[0].lower() != cls.keyword.lower().encode():
            raise cls.exception('Token keyword is not valid')
        if len(auth) != 2:
            raise cls.exception('Token should have 2 part')

        try:
            token = auth[1].decode()
        except UnicodeError:
            raise cls.exception('Unicode Error')

        payload = cls.decode_jwt(token)
        return cls.get_user(payload)

    @classmethod
    def get_user(cls, payload: dict):
        """
        Get UserModel from config else use default UserModel from cls.model
        """
        if (user_id := payload.get('user_id')) is None:
            raise cls.exception('Payload does not have user_id')

        user_model = config['user_model'] or cls.model
        if user := user_model.find_one(id=user_id):
            return user

        raise cls.exception('User not found')

    @classmethod
    def encode_jwt(cls, user_id: int) -> str:
        """Encode JWT from user_id."""
        expire = datetime.utcnow() + config['jwt_config'].life_time
        access_payload = {
            'token_type': 'access',
            'user_id': user_id,
            'exp': expire,
        }
        return jwt.encode(access_payload, config['jwt_config'].key, algorithm=config['jwt_config'].algorithm)

    @classmethod
    def decode_jwt(cls, token: str) -> dict:
        """Decode JWT token to user_id (it can return multiple variable ... )"""
        try:
            return jwt.decode(
                token=token,
                key=config['jwt_config'].key,
                algorithms=[config['jwt_config'].algorithm],
            )
        except JWTError as e:
            raise cls.exception(e)

    @classmethod
    def login(cls, user_id: int) -> str:
        """Alias of encode_jwt()"""
        return cls.encode_jwt(user_id=user_id)

    @staticmethod
    def logout(user_id: int):
        # TODO:
        #   1. Save the token in cache
        #   2. Check logout cached token in authentication()
        pass

    @staticmethod
    def exception(message: str | JWTError, /):
        logger.error(f'JWT Authentication Error: "{message}"')
        return AuthenticationException




import bson
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, field_validator

from panther.configs import config
from panther.db.queries import Query

if config['db_engine'] == 'pantherdb':
    IDType = int
else:
    IDType = str


class Model(PydanticBaseModel, Query):
    id: IDType | None = Field(None, validation_alias='_id')

    @field_validator('id', mode='before')
    def validate_id(cls, value):
        if IDType is str:
            if isinstance(value, str):
                try:
                    bson.ObjectId(value)
                except bson.objectid.InvalidId:
                    raise ValueError('Invalid ObjectId')
            elif not isinstance(value, bson.ObjectId):
                raise ValueError('ObjectId required')
            value = str(value)
        return value

    @property
    def _id(self):
        if IDType is int:
            return self.id
        else:
            return bson.ObjectId(self.id) if self.id else None

    def dict(self, *args, **kwargs):
        return self.model_dump(*args, **kwargs)


class BaseUser(Model):
    first_name: str | None
    last_name: str | None



import random
from pathlib import Path
from unittest import TestCase

import faker
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, field_validator

from panther.configs import config
from panther.db.connection import DBSession
from panther.db.queries.pantherdb_queries import BasePantherDBQuery

f = faker.Faker()


class Model(PydanticBaseModel):
    id: int | None = Field(None, validation_alias='_id')

    @field_validator('id', mode='before')
    def validate_id(cls, value):
        return value

    @property
    def _id(self):
        return self.id


class Book(Model, BasePantherDBQuery):
    name: str
    author: str
    pages_count: int


class TestPantherDB(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        config['db_engine'] = 'pantherdb'

    def setUp(self) -> None:
        DBSession(db_url='pantherdb://database.pdb')

    @classmethod
    def tearDownClass(cls) -> None:
        config['db_engine'] = ''

    def tearDown(self) -> None:
        Path('database.pdb').unlink()

    def test_insert_one(self):
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        book = Book.insert_one(name=name, author=author, pages_count=pages_count)

        self.assertIsInstance(book, Book)
        self.assertEqual(book.id, 1)
        self.assertEqual(book.name, name)
        self.assertEqual(book.pages_count, pages_count)

    def test_find_one(self):
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        # Insert
        created_book = Book.insert_one(name=name, author=author, pages_count=pages_count)
        # FindOne
        book = Book.find_one(name=name, author=author, pages_count=pages_count)

        self.assertIsInstance(book, Book)
        self.assertEqual(book.id, 1)
        self.assertEqual(book.name, name)
        self.assertEqual(book.pages_count, pages_count)
        self.assertEqual(created_book, book)

    def test_find_one_in_many(self):
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        insert_count = random.randint(2, 10)
        # Insert Many
        for _ in range(insert_count):
            Book.insert_one(name=f.name(), author=f.name(), pages_count=random.randint(0, 10))
        created_book = Book.insert_one(name=name, author=author, pages_count=pages_count)
        # FindOne
        book = Book.find_one(name=name, author=author, pages_count=pages_count)

        self.assertIsInstance(book, Book)
        self.assertEqual(book.id, insert_count + 1)
        self.assertEqual(book.name, name)
        self.assertEqual(book.pages_count, pages_count)
        self.assertEqual(created_book, book)

    def test_find(self):
        # Insert Many
        for _ in range(random.randint(2, 10)):
            Book.insert_one(name=f.name(), author=f.name(), pages_count=random.randint(0, 10))

        name = f.name()
        insert_count = random.randint(2, 10)
        for _ in range(insert_count):
            Book.insert_one(name=name, author=f.name(), pages_count=random.randint(0, 10))

        # Find
        books = Book.find(name=name)

        self.assertIsInstance(books, list)
        self.assertEqual(len(books), insert_count)
        book = books[0]
        self.assertIsInstance(book, Book)
        self.assertEqual(book.name, name)

    def test_all(self):
        insert_count = random.randint(2, 10)
        # Insert Many
        for _ in range(insert_count):
            Book.insert_one(name=f.name(), author=f.name(), pages_count=random.randint(0, 10))

        # Find All
        books = Book.find()

        self.assertIsInstance(books, list)
        self.assertEqual(len(books), insert_count)
        book = books[0]
        self.assertIsInstance(book, Book)

    def test_count_all(self):
        insert_count = random.randint(2, 10)
        # Insert Many
        for _ in range(insert_count):
            Book.insert_one(name=f.name(), author=f.name(), pages_count=random.randint(0, 10))

        # Count All
        books_count = Book.count()

        self.assertIsInstance(books_count, int)
        self.assertEqual(books_count, insert_count)

    def test_count_with_filter(self):
        # Insert Many
        for _ in range(random.randint(2, 10)):
            Book.insert_one(name=f.name(), author=f.name(), pages_count=random.randint(0, 10))

        name = f.name()
        insert_count = random.randint(2, 10)
        for _ in range(insert_count):
            Book.insert_one(name=name, author=f.name(), pages_count=random.randint(0, 10))

        # Count
        books_count = Book.count(name=name)

        self.assertIsInstance(books_count, int)
        self.assertEqual(books_count, insert_count)

    def test_delete_one(self):
        # Insert Many
        for _ in range(random.randint(2, 10)):
            Book.insert_one(name=f.name(), author=f.name(), pages_count=random.randint(0, 10))

        name = f.name()
        Book.insert_one(name=name, author=f.name(), pages_count=random.randint(0, 10))

        # Delete One
        is_deleted = Book.delete_one(name=name)

        self.assertIsInstance(is_deleted, bool)
        self.assertTrue(is_deleted)

        # Count Them
        books_count = Book.count(name=name)
        self.assertEqual(books_count, 0)

    def test_delete_one_not_found(self):
        insert_count = random.randint(2, 10)
        # Insert Many
        for _ in range(insert_count):
            Book.insert_one(name=f.name(), author=f.name(), pages_count=random.randint(0, 10))

        # Delete One
        is_deleted = Book.delete_one(name='InvalidName')

        self.assertIsInstance(is_deleted, bool)
        self.assertFalse(is_deleted)

        # Count All
        books_count = Book.count()
        self.assertEqual(books_count, insert_count)
